/*
 * Created on 8 sept. 2004
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package org.sat4j.reader;

import java.io.*;
import java.math.*;
import java.util.*;
import java.util.zip.*;

import org.sat4j.core.*;
import org.sat4j.specs.*;

/**
 * This class is a quick hack to read opb formatted files. The reader skip
 * commented lines (beginning with COMMENT_SYMBOL) and expect constraints of the
 * form: [name :] [[+|-]COEF] [*] [+|-]LIT >=|<=|= DEGREE where COEF and DEGREE
 * are plain integer and LIT is an identifier.
 *
 * @author leberre
 */
public class GoodOPBReader implements Reader, Serializable {

    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long serialVersionUID = 1L;

    private static final String COMMENT_SYMBOL = "*";

    private final ISolver solver;

    private final Map<String, Integer> map = new HashMap<String, Integer>();

    private final IVec<String> decode = new Vec<String>();

    /**
     *
     */
    public GoodOPBReader(final ISolver solver) {
        this.solver = solver;
    }

    @Override
    public IProblem parseInstance(final String filename) throws FileNotFoundException, ParseFormatException,
            IOException,
            ContradictionException {

        if (filename.endsWith(".gz")) {
            parseInstance(new LineNumberReader(
                new InputStreamReader(new GZIPInputStream(new FileInputStream(filename)))));
        } else {
            parseInstance(new LineNumberReader(new FileReader(filename)));
        }
        return this.solver;
    }

    public void parseInstance(final LineNumberReader in) throws ContradictionException, IOException {
        this.solver.reset();
        String line;
        while ((line = in.readLine()) != null) {
            // cannot trim is line is null
            line = line.trim();
            if (line.endsWith(";")) {
                line = line.substring(0, line.length() - 1);
            }
            parseLine(line);
        }
    }

    void parseLine(String line) throws ContradictionException {
        // Skip commented line
        if (line.startsWith(COMMENT_SYMBOL)) {
            return;
        }
        if (line.startsWith("p")) {
            return;
        }
        if (line.startsWith("min:") || line.startsWith("min :"))
         {
            return; // we will use that case later
        }
        if (line.startsWith("max:") || line.startsWith("max :"))
         {
            return; // we will use that case later
        }

        // skip name of constraints:
        final int index = line.indexOf(":");
        if (index != -1) {
            line = line.substring(index + 1);
        }

        final IVecInt lits = new VecInt();
        final IVec<BigInteger> coeffs = new Vec<BigInteger>();
        try (final Scanner lineScanner = new Scanner(line);
                final Scanner stk = lineScanner.useDelimiter("\\s*\\*\\s*|\\s*\\+\\s*|\\s+")) {
            while (stk.hasNext()) {
                String token = stk.next();
                if (token.equals(">=") || token.equals("<=") || token.equals("=")) {
                    assert stk.hasNext();
                    String tok = stk.next();
                    // we need to remove + from the integer
                    if (tok.startsWith("+")) {
                        tok = tok.substring(1);
                    }
                    final BigInteger d = new BigInteger(tok);

                    try {
                        // System.out.println("c constraint: " + line);
                        // System.out.println("c lits: " + lits);
                        // System.out.println("c coeffs: " + coeffs);
                        if (token.equals(">=") || token.equals("=")) {
                            this.solver.addPseudoBoolean(lits, coeffs, true, d);
                        }
                        if (token.equals("<=") || token.equals("=")) {
                            this.solver.addPseudoBoolean(lits, coeffs, false, d);
                        }
                    } catch (final ContradictionException ce) {
                        System.out.println("c inconsistent constraint: " + line);
                        System.out.println("c lits: " + lits);
                        System.out.println("c coeffs: " + coeffs);
                        throw ce;
                    }
                } else {
                    // on est toujours en train de lire la partie gauche de la
                    // contrainte
                    if (token.equals("+")) {
                        assert stk.hasNext();
                        token = stk.next();
                    } else if (token.equals("-")) {
                        assert stk.hasNext();
                        token = token + stk.next();
                    }
                    BigInteger coef;
                    // should contain a coef and a literal
                    try {
                        // we need to remove + from the integer
                        if (token.startsWith("+")) {
                            token = token.substring(1);
                        }
                        coef = new BigInteger(token);
                        assert stk.hasNext();
                        token = stk.next();
                    } catch (final NumberFormatException nfe) {
                        // its only an identifier
                        coef = BigInteger.ONE;
                    }
                    if (token.equals("-") || token.equals("~")) {
                        assert stk.hasNext();
                        token = token + stk.next();
                    }
                    boolean negative = false;
                    if (token.startsWith("+")) {
                        token = token.substring(1);
                    } else if (token.startsWith("-")) {
                        token = token.substring(1);
                        assert coef.equals(BigInteger.ONE);
                        coef = BigInteger.ONE.negate();
                    } else if (token.startsWith("~")) {
                        token = token.substring(1);
                        negative = true;
                    }
                    Integer id = this.map.get(token);
                    if (id == null) {
                        this.map.put(token, id = this.solver.newVar());
                        this.decode.push(token);
                        assert this.decode.size() == id.intValue();
                    }
                    coeffs.push(coef);
                    final int lid = (negative ? -1 : 1) * id.intValue();
                    lits.push(lid);
                    assert coeffs.size() == lits.size();
                }
            }
        }
    }

    @Override
    public String decode(final int[] model) {
        final StringBuffer stb = new StringBuffer();
        for (int i = 0; i < model.length; i++) {
            if (model[i] < 0) {
                stb.append("-");
                stb.append(this.decode.get(-model[i] - 1));
            } else {
                stb.append(this.decode.get(model[i] - 1));
            }
            stb.append(" ");
        }
        return stb.toString();
    }

}
