package org.sat4j.tools;

import java.io.Serializable;

import org.sat4j.core.VecInt;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.IVecInt;

/**
 * Very simple Dimacs array reader. Allow solvers to read the constraints
 * from arrays that effectively contain Dimacs formatted lines
 * (without the terminating 0).
 * 
 * Adaptation of org.sat4j.reader.DimacsReader.
 * 
 * @author dlb
 * @author or
 * @author fuhs
 */
public class DimacsArrayReader implements Serializable {

    private static final long serialVersionUID = 1L;

    protected final ISolver solver;

    public DimacsArrayReader(ISolver solver) {
        this.solver = solver;
    }

    protected boolean handleConstr(int gateType, int output, int[] inputs)
            throws ContradictionException {
        IVecInt literals = new VecInt(inputs);
        solver.addClause(literals);
        return true;
    }

    /**
     * @param gateType
     *             gateType[i] is the type of gate i according to the
     *  Extended Dimacs specs; ignored in DimacsArrayReader, but important
     *  for inheriting classes
     * @param outputs
     *             outputs[i] is the number of the output; ignored in
     *  DimacsArrayReader
     * @param inputs
     *             inputs[i] contains the clauses in DimacsArrayReader; an
     *  overriding class might have it contain the inputs of the current gate
     * @param maxVar
     *             the maximum number of assigned ids  
     * @throws ContradictionException
     *             si le probleme est trivialement inconsitant
     */
    public ISolver parseInstance(int[] gateType,
            int[] outputs, int[][] inputs, int maxVar)
            throws ContradictionException {
        solver.reset();
        solver.newVar(maxVar);
        for (int i = 0; i < outputs.length; ++i) {
            handleConstr(gateType[i], outputs[i], inputs[i]); 
        }
        return solver;
    }

    public String decode(int[] model) {
        StringBuilder stb = new StringBuilder(4*model.length);
        for (int i = 0; i < model.length; i++) {
            stb.append(model[i]);
            stb.append(" ");
        }
        stb.append("0");
        return stb.toString();
    }

    protected ISolver getSolver() {
        return solver;
    }
}
